@@ -14,7 +14,7 @@ use FileHandle          ();
 use vars qw( $VERSION @ISA );
 
 BEGIN {
-    $VERSION = '1.65';
+    $VERSION = '1.66';
 
     require Exporter;
     @ISA = qw( Exporter );
@@ -53,6 +53,9 @@ BEGIN {
               IFA_TEXT_FILE_MASK
               IFA_TEXT_FILE
               IFA_BINARY_FILE
+              ZIP64_AS_NEEDED
+              ZIP64_EOCD
+              ZIP64_HEADERS
               )
         ],
 
@@ -117,9 +120,13 @@ BEGIN {
               DATA_DESCRIPTOR_SIGNATURE
               DATA_DESCRIPTOR_FORMAT
               DATA_DESCRIPTOR_LENGTH
+              DATA_DESCRIPTOR_ZIP64_FORMAT
+              DATA_DESCRIPTOR_ZIP64_LENGTH
 
               DATA_DESCRIPTOR_FORMAT_NO_SIG
               DATA_DESCRIPTOR_LENGTH_NO_SIG
+              DATA_DESCRIPTOR_ZIP64_FORMAT_NO_SIG
+              DATA_DESCRIPTOR_ZIP64_LENGTH_NO_SIG
 
               CENTRAL_DIRECTORY_FILE_HEADER_SIGNATURE
               CENTRAL_DIRECTORY_FILE_HEADER_FORMAT
@@ -137,6 +144,8 @@ BEGIN {
               END_OF_CENTRAL_DIRECTORY_FORMAT
               END_OF_CENTRAL_DIRECTORY_LENGTH
 
+              ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_SIGNATURE_STRING
+              ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIGNATURE_STRING
               END_OF_CENTRAL_DIRECTORY_SIGNATURE_STRING
               )
         ],
@@ -232,6 +241,12 @@ use constant IFA_TEXT_FILE_MASK => 1;
 use constant IFA_TEXT_FILE      => 1;
 use constant IFA_BINARY_FILE    => 0;
 
+# desired zip64 structures for archive creation
+
+use constant ZIP64_AS_NEEDED => 0;
+use constant ZIP64_EOCD      => 1;
+use constant ZIP64_HEADERS   => 2;
+
 # PKZIP file format miscellaneous constants (for internal use only)
 use constant SIGNATURE_FORMAT => "V";
 use constant SIGNATURE_LENGTH => 4;
@@ -242,13 +257,17 @@ use constant LOCAL_FILE_HEADER_FORMAT    => "v3 V4 v2";
 use constant LOCAL_FILE_HEADER_LENGTH    => 26;
 
 # PKZIP docs don't mention the signature, but Info-Zip writes it.
-use constant DATA_DESCRIPTOR_SIGNATURE => 0x08074b50;
-use constant DATA_DESCRIPTOR_FORMAT    => "V3";
-use constant DATA_DESCRIPTOR_LENGTH    => 12;
+use constant DATA_DESCRIPTOR_SIGNATURE    => 0x08074b50;
+use constant DATA_DESCRIPTOR_FORMAT       => "V3";
+use constant DATA_DESCRIPTOR_LENGTH       => 12;
+use constant DATA_DESCRIPTOR_ZIP64_FORMAT => "L< Q<2";
+use constant DATA_DESCRIPTOR_ZIP64_LENGTH => 20;
 
 # but the signature is apparently optional.
-use constant DATA_DESCRIPTOR_FORMAT_NO_SIG => "V2";
-use constant DATA_DESCRIPTOR_LENGTH_NO_SIG => 8;
+use constant DATA_DESCRIPTOR_FORMAT_NO_SIG       => "V2";
+use constant DATA_DESCRIPTOR_LENGTH_NO_SIG       => 8;
+use constant DATA_DESCRIPTOR_ZIP64_FORMAT_NO_SIG => "Q<2";
+use constant DATA_DESCRIPTOR_ZIP64_LENGTH_NO_SIG => 16;
 
 use constant CENTRAL_DIRECTORY_FILE_HEADER_SIGNATURE => 0x02014b50;
 use constant CENTRAL_DIRECTORY_FILE_HEADER_FORMAT    => "C2 v3 V4 v5 V2";
@@ -256,13 +275,16 @@ use constant CENTRAL_DIRECTORY_FILE_HEADER_LENGTH    => 42;
 
 # zip64 support
 use constant ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_SIGNATURE => 0x06064b50;
-use constant ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_FORMAT => 0;
-use constant ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_LENGTH => 0;
+use constant ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_SIGNATURE_STRING =>
+  pack("V", ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_SIGNATURE);
+use constant ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_FORMAT => "Q< S<2 L<2 Q<4";
+use constant ZIP64_END_OF_CENTRAL_DIRECTORY_RECORD_LENGTH => 52;
 
 use constant ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIGNATURE => 0x07064b50;
-use constant ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_FORMAT => 0;
-use constant ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_LENGTH => 0;
-
+use constant ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIGNATURE_STRING =>
+  pack("V", ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIGNATURE);
+use constant ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_FORMAT => "L< Q< L<";
+use constant ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_LENGTH => 16;
 
 use constant END_OF_CENTRAL_DIRECTORY_SIGNATURE => 0x06054b50;
 use constant END_OF_CENTRAL_DIRECTORY_SIGNATURE_STRING =>
@@ -462,13 +484,20 @@ sub _newFileHandle {
 
 # Returns next signature from given file handle, leaves
 # file handle positioned afterwards.
+#
 # In list context, returns ($status, $signature)
-# ( $status, $signature) = _readSignature( $fh, $fileName );
-
+# ( $status, $signature ) = _readSignature( $fh, $fileName );
+#
+# This function returns one of AZ_OK, AZ_IO_ERROR, or
+# AZ_FORMAT_ERROR and calls the respective error handlers in the
+# latter two cases.  If optional $noFormatError is true, it does
+# not call the error handler on format error, but only returns
+# AZ_FORMAT_ERROR.
 sub _readSignature {
     my $fh                = shift;
     my $fileName          = shift;
     my $expectedSignature = shift;    # optional
+    my $noFormatError     = shift;    # optional
 
     my $signatureData;
     my $bytesRead = $fh->read($signatureData, SIGNATURE_LENGTH);
@@ -490,12 +519,17 @@ sub _readSignature {
             && $signature != ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIGNATURE
         )
       ) {
-        my $errmsg = sprintf("bad signature: 0x%08x", $signature);
-        if (_isSeekable($fh)) {
-            $errmsg .= sprintf(" at offset %d", $fh->tell() - SIGNATURE_LENGTH);
-        }
+        if (! $noFormatError ) {
+            my $errmsg = sprintf("bad signature: 0x%08x", $signature);
+            if (_isSeekable($fh)) {
+                $errmsg .= sprintf(" at offset %d", $fh->tell() - SIGNATURE_LENGTH);
+            }
 
-        $status = _formatError("$errmsg in file $fileName");
+            $status = _formatError("$errmsg in file $fileName");
+        }
+        else {
+            $status = AZ_FORMAT_ERROR;
+        }
     }
 
     return ($status, $signature);
@@ -722,6 +756,7 @@ COMPRESSION_STORED COMPRESSION_DEFLATED IFA_TEXT_FILE_MASK
 IFA_TEXT_FILE IFA_BINARY_FILE COMPRESSION_LEVEL_NONE
 COMPRESSION_LEVEL_DEFAULT COMPRESSION_LEVEL_FASTEST
 COMPRESSION_LEVEL_BEST_COMPRESSION
+ZIP64_AS_NEEDED ZIP64_EOCD ZIP64_HEADERS
 
 =item :MISC_CONSTANTS
 
@@ -1026,6 +1061,26 @@ members in scalar context.
     # or
     my $numberOfTextFiles = $zip->membersMatching( '.*\.txt' );
 
+=item zip64()
+
+Returns whether the previous read or write of the archive has
+been done in zip64 format.
+
+=item desiredZip64Mode()
+
+Gets or sets which parts of the archive should be written in
+zip64 format: All parts as needed (ZIP64_AS_NEEDED), the default,
+force writing the zip64 end of central directory record
+(ZIP64_EOCD), force writing the zip64 EOCD record and all headers
+in zip64 format (ZIP64_HEADERS).
+
+=item versionMadeBy()
+
+=item versionNeededToExtract()
+
+Gets the fields from the zip64 end of central directory
+record. These are always 0 if the archive is not in zip64 format.
+
 =item diskNumber()
 
 Return the disk that I start on. Not used for writing zips,
@@ -1690,12 +1745,30 @@ Returns undef on error.
 
 =back
 
-=head2 Member Simple accessors
+=head2 Member Simple Accessors
 
 These methods get (and/or set) member attribute values.
 
+The zip64 format requires parts of the member data to be stored
+in the so-called extra fields.  You cannot get nor set this zip64
+data through the extra field accessors described in this section.
+In fact, the low-level member methods ensure that the zip64 data
+in the extra fields is handled completely transparently and
+invisibly to the user when members are read or written.
+
 =over 4
 
+=item zip64()
+
+Returns whether the previous read or write of the member has been
+done in zip64 format.
+
+=item desiredZip64Mode()
+
+Gets or sets whether the member's headers should be written in
+zip64 format: As needed (ZIP64_AS_NEEDED), the default, or always
+(ZIP64_HEADERS).
+
 =item versionMadeBy()
 
 Gets the field from the member header.
@@ -1811,17 +1884,20 @@ comparisons.
 =item localExtraField( [ { field => $newField } ] )
 
 Gets or sets the extra field that was read from the local
-header. This is not set for a member from a zip file until
-after the member has been written out. The extra field must
-be in the proper format.
+header. The extra field must be in the proper format.  If it is
+not or if the new field contains data related to the zip64
+format, this method does not modify the extra field and returns
+AZ_FORMAT_ERROR, otherwise it returns AZ_OK.
 
 =item cdExtraField( [ $newField ] )
 
 =item cdExtraField( [ { field => $newField } ] )
 
 Gets or sets the extra field that was read from the central
-directory header. The extra field must be in the proper
-format.
+directory header. The extra field must be in the proper format.
+If it is not or if the new field contains data related to the
+zip64 format, this method does not modify the extra field and
+returns AZ_FORMAT_ERROR, otherwise it returns AZ_OK.
 
 =item extraFields()
 
@@ -2007,6 +2083,10 @@ change the class of the member):
 Extract (and uncompress, if necessary) the member's contents
 to the given file handle. Return AZ_OK on success.
 
+For members representing symbolic links, pass the name of the
+symbolic link as file handle. Ensure that all directories in the
+path to the symbolic link already exist.
+
 =back
 
 =head1 Archive::Zip::FileMember methods
