@@ -4,7 +4,7 @@ use strict;
 use vars qw( $VERSION @ISA );
 
 BEGIN {
-    $VERSION = '1.65';
+    $VERSION = '1.66';
     @ISA     = qw( Archive::Zip::Member );
 }
 
@@ -49,7 +49,7 @@ sub contents {
           length($self->{'contents'});
         $self->{'compressionMethod'} = COMPRESSION_STORED;
     }
-    return $self->{'contents'};
+    return wantarray ? ($self->{'contents'}, AZ_OK) : $self->{'contents'};
 }
 
 # Return bytes read. Note that first parameter is a ref to a buffer.
